# Created by Noé Cruz | Zurckz 22 at 12/11/2022
# See https://www.linkedin.com/in/zurckz
from zpy.utils.functions import Result, Ok, Failure


def build(wrap_ok: bool = False):
    def decorator(function):
        def wrapper(*args, **kwargs):
            try:
                result = function(*args, **kwargs)
                if wrap_ok:
                    return Ok.of(result)
                return result
            except Exception as e:
                return Failure.of(e)

        return wrapper

    return decorator



# Created by Noé Cruz | Zurckz 22 at 24/07/2022
# See https://www.linkedin.com/in/zurckz
from typing import Optional, Any, List

from zpy.logger import ZLogger
from zpy.utils.values import if_null_get
from zpy.utils import get_env
from logging import getLevelName, INFO, WARN, ERROR, DEBUG
from zpy.utils.collections import walk_and_apply
from copy import copy
from zpy.app import zapp_context as ctx

sensitive_data_filters = []
main_logger: Optional[ZLogger] = None
main_logger_app: Optional[str] = None


def init_mask_meta_keys_filter(keys: List[str] = None):
    keys = ['pass', 'password', 'bank', 'pwd', 'contraseña', 'contra', 'priv', 'banc', 'secret', 'clave']

    def mask_meta_keys_filter(log: dict) -> dict:
        def sensitive_data_masker(key, value) -> Any:
            for sk in keys:
                if isinstance(key, str) and sk in copy(key).lower():
                    return f"******{str(value)[-2:]}"
            return value

        if 'meta' in log and isinstance(log['meta'], dict):
            walk_and_apply(log['meta'], sensitive_data_masker)
        return log

    return mask_meta_keys_filter


class StructuredLog(object):

    @staticmethod
    def add_sensitive_filter(log_filter):
        global sensitive_data_filters
        sensitive_data_filters.append(log_filter)

    @staticmethod
    def configure_logger(logger, app_name: str):
        global main_logger
        main_logger = logger
        global main_logger_app
        main_logger_app = app_name

    def __init__(self, flow: str, message: str, app: str = None, description: str = None, level: int = INFO,
                 meta: Optional[dict] = None, **kwargs) -> None:
        metadata = if_null_get(meta, {})
        metadata.update(kwargs)
        self.log: dict = {
            'application': if_null_get(app, get_env("APP_NAME", "Unknown")),
            'flow': flow,
            'message': message,
            'description': description,
            'level': getLevelName(level),
            'meta': metadata
        }
        if sensitive_data_filters:
            self.log = [sf(self.log) for sf in sensitive_data_filters][0]

    def __str__(self):
        return f'{self.log}'

    @staticmethod
    def log_info(flow: str, message: str, description: str = None, meta: Optional[dict] = None,
                 **kwargs) -> 'StructuredLog':
        log = StructuredLog(flow, message, main_logger_app, description, INFO, meta, **kwargs)
        if_null_get(main_logger, ctx().logger).info(log)
        return log

    @staticmethod
    def log_warn(flow: str, message: str, description: str = None, meta: Optional[dict] = None,
                 **kwargs) -> 'StructuredLog':
        log = StructuredLog(flow, message, main_logger_app, description, WARN, meta, **kwargs)
        if_null_get(main_logger, ctx().logger).warn(log)
        return log

    @staticmethod
    def log_error(flow: str, message: str, description: str = None, meta: Optional[dict] = None,
                  **kwargs) -> 'StructuredLog':
        log = StructuredLog(flow, message, main_logger_app, description, ERROR, meta, **kwargs)
        if_null_get(main_logger, ctx().logger).err(log)
        return log



# Created by Noé Cruz | Zurckz 22 at 24/09/2022
# See https://www.linkedin.com/in/zurckz
from abc import ABC, abstractmethod
from typing import Any


class UnknownKeyException(Exception):
    def __init__(self, key: str, message: str = "The provided key does not exist"):
        super().__init__(message)
        self.key = key


class Storable(ABC):
    """
    Simple Storable Definition
    """

    @abstractmethod
    def clear(self):
        """
        Clear all data stored
        @return:
        """
        pass

    @abstractmethod
    def set(self, key: Any, value: Any):
        """
        Store some data by key
        @param key: key for identifier value
        @param value: value to store
        @return: None
        """
        pass

    @abstractmethod
    def get(self, key: Any) -> None:
        """
        Retrieve data by key
        @param key:
        @return:
        """
        pass

    @abstractmethod
    def get_or(self, key: Any, default: Any = None) -> None:
        """
        Retrieve data by key
        @param default:
        @param key:
        @return:
        """
        pass



from typing import Any, Callable, List, Optional, Tuple, Union
from functools import reduce
import logging
import time
import inspect

from zpy.logger import ZLogger
from zpy.utils.values import if_null_get
from zpy.app import zapp_context as ctx


def safely_exec(callable_fn: Callable, args=None) -> Optional[Any]:
    """
    Execute provided function in try:except block
    @param callable_fn:
    @param args:
    @return: value returned of execution or none
    """
    if args is None:
        args = []
    try:
        return callable_fn(*args)
    except Exception as e:
        logging.exception(e)
    return None


def safe_exec_wrapper(target: Callable, args=None, kwargs: dict = None,
                      msg: str = None,
                      notifier: Optional[Callable] = None,
                      default_ret=None, throw_ex: bool = False):
    logg = ctx().logger
    if args is None:
        args = []
    if not kwargs:
        kwargs = {}
    try:
        return target(*args, **kwargs)
    except Exception as e:
        msg = if_null_get(msg, f"An error occurred while try execute: {target.__name__} with: {args} and {kwargs}.")
        logg.exception(msg)
        if notifier:
            notifier(f"Fatal: {msg}\n{str(e)}")
        if throw_ex:
            raise
    return default_ret


def exec_if_nt_null(callable_fn: Callable, args: Optional[List[Any]] = None) -> object:
    """
    Execute function if args not null
    """
    if args is None:
        args = []
    for arg in args:
        if arg is None:
            return False
    return callable_fn(*args)


def safely_exec_with(callable_fn: Callable, default_value: Any = None, args=None) -> Optional[Any]:
    """
    Execute provided function in try:except block
    @param default_value:
    @param callable_fn:
    @param args:
    @return: value returned of execution or none
    """
    if args is None:
        args = []
    try:
        return callable_fn(*args)
    except Exception as e:
        logging.exception(e)
    return default_value


def get_class_that_defined_method(meth):
    for cls in inspect.getmro(meth.im_self):
        if meth.__name__ in cls.__dict__:
            return cls
    return None


def timeit(msg: str = None):
    """
    Time execution logger.
    @param msg:
    @return:
    """

    def _timeit_(method):
        def timed(*args, **kw):
            logg = ctx().logger
            ts = time.time()
            result = method(*args, **kw)
            te = time.time()

            method_name = ('{} -> {}'.format(method.__module__, method.__name__)) if not msg else msg
            if 'log_time' in kw:
                name = kw.get('log_name', method.__name__.upper())
                kw['log_time'][name] = int((te - ts) * 1000)
            else:
                logg.info(f"Time Execution: {method_name} :: {(te - ts) * 1000:2.2f} ms.")

            return result

        return timed

    return _timeit_


def safe(default: Any = None, notifier: Any = None, logger: ZLogger = None, err_msg: str = None):
    """
    Safe execution
    @param logger:
    @param err_msg:
    @param notifier: Error notifier
    @param default: if execution fail return default value
    @return:
    """

    def __safe_execution__(func):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                method_name = ('{} -> {}'.format(func.__module__, func.__name__)) if not err_msg else err_msg
                error_message = f"An error occurred while executing {method_name}. Msg: {str(e)}"
                if notifier:
                    notifier(error_message)
                if logger:
                    logger.err(error_message, exc_info=e)
            return default

        return wrapper

    return __safe_execution__


def fn_composite(*func):
    """
    Function composition
    @param func: functions
    @return: composition
    """

    def compose(f, g):
        return lambda x: f(g(x))

    return reduce(compose, func, lambda x: x)


def if_then(value: bool, function: Callable, args: Union[List, Tuple] = None, default_value: Any = None):
    """
    If value provided is true, then execute function provided as param.
    @param value:
    @param function:
    @param args:
    @param default_value:
    @return:
    """
    if value:
        return function(*if_null_get(args, []))
    return default_value


def if_get(value: bool, if_value: Any = None, else_value: Any = None, strict=False):
    """
    If value provided is true, then return values provided as param.
    @param strict: Force boolean value to compare
    @param if_value: value returned if value is true
    @param else_value: value returned if value is false
    @param value: value to evaluate
    @return: Any
    """

    if strict is True and value is True:
        return if_value
    if strict is False and value:
        return if_value
    return else_value


def if_not_then(value: bool, function: Callable, args: Union[List, Tuple] = None, default_value: Any = None):
    """
    If value provided is true, then execute function provided as param.
    @param value:
    @param function:
    @param args:
    @param default_value:
    @return:
    """

    if not value:
        return function(*if_null_get(args, []))

    return default_value


def if_else_then(value: bool, if_func: Callable = None, if_args: Union[List, Tuple] = None, else_func: Callable = None,
                 else_args: Union[List, Tuple] = None, if_value: Any = None, else_value: Any = None):
    """
    If value provided is true, then execute if_function provided as param otherwise execute else_function
    @param else_value:
    @param if_value:
    @param else_args:
    @param else_func:
    @param if_args:
    @param value:
    @param if_func:

    @return:
    """

    if value:
        if if_func:
            return if_func(*if_null_get(if_args, []))
        return if_value
    if else_func:
        return else_func(*if_null_get(else_args, []))
    return else_value


class Maybe(object):
    """

    """

    def __init__(self, value) -> None:
        self.value = value
        self.exception = None

    @classmethod
    def unit(cls, value):
        return cls(value)

    def __getattr__(self, item):
        _field = getattr(self.value, item)
        if not callable(_field):
            return self.bind(lambda _: _field)
        return lambda *args, **kwargs: self.bind(lambda _: _field(*args, **kwargs))

    def apply(self, fn):
        if fn is None:
            return self
        fn(self.value)
        return self

    def bind(self, fn):
        if self.value is None:
            return self
        result = fn(self.value)
        if isinstance(result, Maybe):
            return result

        return Maybe.unit(result)

    def safe_bind(self, fn):
        try:
            return self.bind(fn)
        except Exception as e:
            self.exception = e
            return self

    def safe_apply(self, fn):
        try:
            return self.bind(fn)
        except Exception as e:
            self.exception = e
            return self

    def optional_bind(self, fn):
        return self.safe_bind(fn)

    def optional_apply(self, fn):
        return self.safe_apply(fn)

    def unwrap(self) -> Any:
        if self.exception:
            raise self.exception

        return self.value

    def value_or(self, default_value: Any) -> Any:
        return if_null_get(self.value, default_value)



from abc import abstractmethod
from typing import List, Tuple, Union, Any, TypeVar

from marshmallow.exceptions import ValidationError

from zpy.api.http.errors import BadRequest, ZHttpError
from zpy.utils.objects import ZObjectModel

R = TypeVar('R')


class ZRequest:

    @abstractmethod
    def verify(self, *args, **kwargs):
        """
        Execute extra validations over data
        @param args:
        @param kwargs:
        @return:
        """
        pass

    @abstractmethod
    def patch(self, *args, **kwargs):
        """
        Update properties values
        @param args:
        @param kwargs:
        @return:
        """
        pass


def query(queries: dict, name: str, default: Any = None, raise_errors=True, type_data: Any = None):
    """
    Extract query param from dict
    @param queries:
    @param name:
    @param default:
    @param raise_errors:
    @param type_data:
    @return:
    """

    def raise_or_return(msg: str):
        if raise_errors:
            raise BadRequest(reason=msg)
        return default

    if not queries or name not in queries:
        return raise_or_return(f"Missing query param: {name}")
    value = queries.get(name, None)
    if not value:
        return raise_or_return(f"Missing value query param: {name}")
    if type_data and not isinstance(value, type_data):
        return raise_or_return(
            f"Unexpected type data for query param value: {name}. Expected type: {type_data.__name__}")
    return value


def parse_request(
        request: dict, model: Union[R, Any], raise_err: bool = True
) -> Tuple[Union[R, Any], ZHttpError]:
    """
    Parse and validate request according model specification.

    @param request:
    @param model:
    @param raise_err:
    @return:
    """

    model_result: Union[List, ZObjectModel, Any] = None
    errors: Union[List, None, BadRequest] = None
    try:
        if request is None or len(request.items()) == 0:
            error = BadRequest(
                "The request was not provided, validation request error",
                f"Missing fields {model().__missing_fields__} not provided",
            )
            if raise_err:
                raise error
            return None, error
        model_result = model(**request)
    except ValidationError as e:
        model_result = e.valid_data
        # if isinstance(e.messages, Dict):
        #     errors = [e.messages]
        # else:
        #     errors = e.messages
        errors = BadRequest(None, f"{e.messages}")
        if raise_err:
            raise errors
    return model_result, errors



# Created by Noé Cruz | Zurckz 22 at 05/03/2022
# See https://www.linkedin.com/in/zurckz
import logging
from typing import Optional
from zpy.utils import get_env
from zpy.utils.values import if_null_get
from zpy.logger import zL, ZLFormat, ZLogger


class ZContext(object):

    def __init__(self, logger: Optional[ZLogger] = None, app_name: str = None, env: str = None, flow: str = None):
        self.logger = logger
        self.app_name = if_null_get(app_name, get_env("APP_NAME", "Z-API"))
        self.environment = if_null_get(env, get_env("ENVIRONMENT", "dev"))
        self.main_flow = if_null_get(flow, get_env("FLOW", ""))
        self.flow = self.main_flow
        if not logger:
            self.logger = zL(self.app_name, logging.INFO, ZLFormat.LM)

    def release_logger(self, name: str = "Z-API"):
        self.logger.release(name)

    def update_flow(self, flow: str):
        self.flow = f'{self.flow}#{flow}'

    def get_flow(self):
        return self.flow

    def reset_flow(self):
        self.flow = self.main_flow


context: Optional[ZContext] = None


def setup_context(ctx: Optional[ZContext] = None) -> ZContext:
    global context
    if ctx:
        context = ctx
    if not ctx:
        context = ZContext()
    return context


def zapp_context() -> ZContext:
    global context
    if not context:
        return setup_context()
    return context



# Created by Noé Cruz | Zurckz 22 at 04/08/2022
# See https://www.linkedin.com/in/zurckz
from datetime import datetime


class EventContext:
    aws_request_id: str

    def __init__(self):
        self.aws_request_id = str(datetime.utcnow().timestamp())



# Created by Noé Cruz | Zurckz 22 at 24/09/2022
# See https://www.linkedin.com/in/zurckz
from typing import Any

from zpy.containers.Storable import Storable, UnknownKeyException


class MemoryCacheSystem(Storable):
    """
    Simple cache implementation.
    TODO: Strategies for expiring cache
    """

    def __init__(self):
        self.values = {}

    def clear(self):
        self.values.clear()

    def set(self, key: Any, value: Any):
        self.values[key] = value

    def get(self, key: Any) -> None:
        if key in self.values:
            return self.values[key]
        raise UnknownKeyException(key=key)

    def get_or(self, key: Any, default: Any = None) -> Any:
        return self.values.get(key, default)



from abc import ABC, abstractmethod
from typing import Generic, TypeVar, List, Union, Any, Dict
from copy import copy
from zpy.utils.funcs import Maybe, if_get
from zpy.utils.values import if_null_get
from zpy.app import zapp_context as ctx

T = TypeVar("T")
S = TypeVar("S")


class UCMeta:
    def __init__(self, identifier: str, key_identifier: str = None) -> None:
        self.id = identifier
        if key_identifier:
            setattr(self, key_identifier, identifier)


class Selectable:
    def __init__(self, identifier: str, key_identifier: str = None, weight: float = 1.0) -> None:
        """

        @param identifier: value of selector based. E.g. UserCreator
        @param key_identifier: key of selector based. E.g. action
        """
        if not key_identifier:
            key_identifier = 'id'
        self.weight = weight
        self.name = identifier
        self.identifiers: Dict[str, Dict[str, str]] = {
            key_identifier: {"value": identifier, "weight": weight}
        }

    def configure_for_all(self, weight: float = 2.0):
        self.identifiers['*'] = {"value": '*', "weight": weight}

    def configure(self, uc_identifier: Any, weight: float, key_identifier: Any = 'id'):
        self.identifiers[key_identifier] = {
            "value": uc_identifier,
            "weight": weight
        }


class UseCase(ABC, Generic[T, S]):
    def __init__(self, name: Any = None):
        self.name = name

    def before(self):
        pass

    @abstractmethod
    def execute(self, payload: T, *args, **kwargs) -> S:
        """Execute use case"""
        pass

    def after(self):
        pass


class UseCaseSelector(UseCase, UCMeta):

    def __init__(self, use_cases: List[Union[UseCase, UCMeta, Any]], action_keys: List[str] = None,
                 key_uc_identifier: str = 'id', selector_id='default', payload_keys: List[str] = None):
        UCMeta.__init__(self, identifier=selector_id)
        self.cases = {getattr(x, key_uc_identifier): x for x in use_cases}
        self.action_keys = if_null_get(action_keys, ['action'])
        self.key_identifier = key_uc_identifier
        self.payload_keys = if_null_get(payload_keys, ['payload'])

    def execute(self, data: dict, context: Any = None, *args, **kwargs) -> dict:
        action = None
        for key_action in self.action_keys:
            if key_action in data:
                action = key_action
                break
        if action is None:
            raise ValueError(f'Request provided is malformed. Missing {action} key!')

        operation: Union[UseCase, UCMeta] = self.cases.get(data[action], None)

        if not operation:
            raise ValueError(f"Use case for action: {data['action']} not registered in selector.")

        payload_key = None
        for pk in self.payload_keys:
            if pk in data:
                payload_key = pk
                break

        payload = data.get(payload_key, data)
        return operation.execute(payload, context=context)


class CaseSelector(UseCase, Selectable):

    def __init__(self, use_cases: List[Union[UseCase, Selectable, Any]], action_keys: List[str] = None,
                 key_uc_identifier: str = 'id', selector_id='default', payload_keys: List[str] = None,
                 payload_mutation: bool = False):
        Selectable.__init__(self, identifier=selector_id, key_identifier=key_uc_identifier)
        self.action_keys = if_null_get(action_keys, ['action'])
        self.payload_keys = if_null_get(payload_keys, ['payload'])
        self.allow_payload_mutation = payload_mutation
        self.on_before = None
        self.on_after = None
        self.cases = Maybe(use_cases) \
            .bind(self.__group_cases) \
            .bind(self.__sort_cases) \
            .value

    @staticmethod
    def __group_cases(cases: List[Union[UseCase, Selectable, Any]]):
        group_to_sort = {}
        for case in cases:
            for k, v in case.identifiers.items():
                if v['value'] not in group_to_sort:
                    group_to_sort[v['value']] = {
                        v['weight']: case
                    }
                    continue
                group_to_sort[v['value']][v['weight']] = case

        return group_to_sort

    @staticmethod
    def __sort_cases(cases: Dict[str, Dict[Any, Union[UseCase, Selectable, Any]]]):
        return {c: [x[1] for x in sorted(cases[c].items())] for c in cases}

    def execute(self, data: dict, context: Any = None, *args, **kwargs) -> List[Any]:
        action = None
        for key_action in self.action_keys:
            if key_action in data:
                action = key_action
                break

        if action is None:
            raise ValueError(f'Request provided is malformed. Missing {action} key!')

        cases_to_execute: List[Union[UseCase, Selectable]] = self.cases.get(data[action], [])
        cases_to_execute = sorted(cases_to_execute + self.cases.get('*', []), key=lambda x: x.weight)

        if not cases_to_execute:
            raise ValueError(f"Use case for action: {data['action']} not registered in selector.")

        payload_key = None
        for pk in self.payload_keys:
            if pk in data:
                payload_key = pk
                break

        payload = data.get(payload_key, data)
        results = []
        if self.on_before:
            self.on_before()
        for x_case in cases_to_execute:
            ctx().logger.info(f'⚡ Running case: {x_case.name}...')
            result = x_case.execute(if_get(self.allow_payload_mutation, payload, copy(payload)), context=context)
            if isinstance(result, dict):
                result['event'] = data[action]
            results.append(result)
        if self.on_after:
            self.on_after()
        return results



import os
from typing import List, Dict
from zpy.utils.files import get_files_in, file_content_updater
from zpy.cli.agregations import add_config, add_context, add_use_case, add_route_event_manager
from zpy.cli.scaffolding import scaffolding_dispatcher

import click


@click.group()
def cli():
    """This is an example script to learn Click."""
    ...


@cli.command()
@click.option('--project', '-p', required=True, type=str, help='Project name')
@click.option('--verbose', '-v', required=False, is_flag=True, default=False, help='Show logs')
@click.option('--description', '-d', required=False, type=str, help='Project description',
              default="This is very boring haha xd...")
@click.option('--developer', '-dev', required=False, type=str, help='Author', default="Zurck'z 2022")
@click.option('--context', '-c', required=False, type=str, help='Context name', default="Users")
@click.option('--use_case', '-uc', required=False, type=str, help='Use Case Name', default="GetUser")
@click.option('--developer_url', '-dev-url', required=False, type=str, help='Author url',
              default="https://www.linkedin.com/in/zurckz/")
@click.option('--open_project', '-op', required=False, is_flag=True, default=False, help='Open project')
@click.option('--only_content', '-oc', required=False, is_flag=True, default=False,
              help='Create only structure project')
def make(project, verbose, description, developer, context, use_case, developer_url, open_project,
         only_content):
    """
    Create new project
    \f
    @param project:
    @param verbose:
    @param description:
    @param developer:
    @param context:
    @param use_case:
    @param developer_url:
    @param open_project:
    @param only_content:
    @return:
    """
    in_data: List[Dict[str, str]] = [{"value": project, "title": "Project name:"},
                                     {"value": description, "title": "Project description:"},
                                     {"value": developer, "title": "Author:"},
                                     {"value": developer_url, "title": "Url Author:"},
                                     {"value": context.lower(), "title": "Context:"},
                                     {"value": use_case, "title": "Use Case:"},
                                     {"value": context.title(), "title": "Repository"},
                                     {"value": open_project, "title": "Open Project?:"},
                                     {"value": only_content, "title": "Only Content?:"},
                                     None
                                     ]
    if verbose is True:
        click.echo(f'\N{grinning face} - Initializing for create new project...')
        for data in in_data:
            if data is not None:
                click.echo(f'\N{grinning face} - With {data["title"]} {data["value"]}')

    scaffolding_dispatcher(in_data, verbose, click)


@cli.group()
def add():
    """
    Add new context, case or event mappers
    \f
    @return:
    """
    ...


@add.command()
@click.option('--name', '-n', required=True, type=str, is_flag=False, help='Bounded Context name')
@click.option('--use_case', '-uc', required=False, type=str, help='Use Case Name', default="GetUser")
@click.option('--force', '-f', required=False, type=bool, is_flag=True, help='Force creation, skipping project',
              default=False)
def context(name: str, use_case: str, force: bool):
    """
    Add new context to project.
    \f
    @param name:
    @param use_case:
    @param force:
    @return:
    """
    add_context(name, use_case, force)


@add.command()
@click.option('--name', '-n', required=True, type=str, is_flag=False, help='Use case name')
@click.option('--context', '-c', required=True, type=str, is_flag=False,
              help='Context name where use case will be created')
# @click.option('--repository', '-r', required=False, type=str, help='Repository dependency', default="AwesomeRepository")
def case(name: str, context: str):
    """
    Add new use case for specific context
    \f
    @param name:
    @param context:
    @return:
    """
    add_use_case(name, context)


@add.command(name='event-mapper')
@click.option('--force', '-f', required=False, type=bool, is_flag=True,
              help='Force creation, if exist the file will be replaced',
              default=False)
def event_mapper_manager(force: bool):
    """
    Create basic event mapper manager configuration
    \f
    @param force:
    @param name:
    @param context:
    @return:
    """
    add_route_event_manager(force)


@add.command()
@click.option('--database', '-db', required=True, type=bool, is_flag=True, help='Database configuration')
def config(database: bool):
    """
    Add basic database configuration
    \f
    @param database:
    @return:
    """
    add_config(database)


@cli.command()
def drop():
    """
    Not available yet.
    \f
    @return:
    """
    click.echo('Dropping...')


@cli.command()
@click.option('--directory', '-d', required=False, type=str, is_flag=False, help='Directory', default='.')
@click.option('--extension', '-ex', required=True, type=str, help='File type')
@click.option('--find', '-fs', required=True, type=str, help='Find substring for replace')
@click.option('--new', '-ns', required=False, type=str, help='New substring. Use: [@EMPTY] for empty str',
              default="[@EMPTY]")
def content_replace(directory: str, extension: str, find: str, new: str = '[@EMPTY]'):
    """
    Replace specific string in files with provided extension
    \f
    @param directory:
    @param extension:
    @param find:
    @param new:
    @return:
    """
    if directory == '.':
        directory = os.getcwd()
    if new in ['[@EMPTY]']:
        new = ''
    click.echo('Starting files content replacement...')
    click.echo(f'Directory: {directory}')
    click.echo(f'File Filtered by: {extension}')
    click.echo(f'Value to replace: \'{find}\'')
    click.echo(f'New value: \'{new}\'')

    def custom_mutator(line: str) -> str:
        if line and "from" in line and "src." in line:
            return line.replace(find, new)
        return line

    for file in get_files_in(directory, extension, True):
        file_content_updater(file, find=find, replaced=new)



# Templates
adds = {
    "di.py": {
        "type": "file",
        "module": "db-oracle",
        "path": "src/di.py",
        "code": {
            "imports": [
                "from zpy.utils import get_env_or_throw as var\n",
                "from zdb.oracle import ZOracle, ZDBConfig\n"
            ],
            "blocks": [
                "\n",
                "db_config: ZDBConfig = ZDBConfig(\"DB_USER\", \"DB_PASSWORD\", \"DB_NAME\",\"DB_HOST\", 1521,service=\"XE\")\n"
                "db_mngr: ZOracle = ZOracle.setup_of(db_config)\n"
                "\n",
                "if var('ENVIRONMENT') == 'local':\n",
                "   # Setup only the environment is local.\n",
                "   db_mngr.init_local_client(path= var('ORACLE_CLIENT_PATH'))\n",
                "\n"
            ]
        }
    },
    "context_imports": {
        "type": "file",
        "module": "db-oracle",
        "path": "src/di.py",
        "code": {
            "imports": [
                "\n"
                "# Imports of @context_name context 📦\n"
                "from contexts.@context_name import @usecase_class, @repository_name, @repository_impl\n"
            ],
            "blocks": [
                "\n# Creation of instances of use cases, repositories etc. of the context: @context_name 📦\n",
                "@context_name_repository: @repository_name = @repository_impl()\n"
                "@usecase_var_uc: UseCase[Any, Any] = @usecase_class(@context_name_repository)\n"
                "\n"
            ]
        }
    },
    "use-cases": {
        "type": "file",
        "module": "cli",
        "path": "src/di.py",
        "code": {
            "imports": [
                "\n"
                "# Import use case of @context_name context 📦\n"
                "from src.contexts.@context_name.application.@usecase_file import @usecase_class\n"
            ],
            "blocks": [
                "\n# Creation instance of use cases of the context: @context_name 📦\n",
                "@usecase_var: UseCase[Any, Any] = @usecase_class(@context_name_repository)\n"
                "\n"
            ]
        }
    },
    "routes-event-mappers": {
        "type": "file",
        "module": "cli",
        "path": "src/api/config.py",
        "code": {
            "imports": [
                "# 🛸 Generated by zPy"
                "\n"
                "from zpy.api.flask.cloud_handlers import RouteEventMapperManager, RouteEventMapper as Route\n",
                "from src.contexts.users.infrastructure.mutators import RatingDetailResponseMutator # ⚠️ You need to define this\n"
            ],
            "blocks": [
                "\n# Event Route Manager configurations",
                """\n# 💡 Remove any configuration you don't need for each route
# 💡 If you don't use initializer remove param
event_mapper_manager = RouteEventMapperManager(initializer=lambda e, c: e) \\
    .add_meta('your_shared_dependency', instance_or_value) \\
    .add(Route(route="/users/{id}/ratings/{rating-id}")
         .for_params(ParamsDecrypterMapper(path_params_keys=['id', 'rating-id']))
         .for_request(BodyDecrypterMapper())
         .for_response(BodyEncrypterMapper())
         .with_meta('request_model', UpdateRatingRequestMutator)
         .with_meta('response_model', RatingDetailResponseMutator)
         .with_patch()) \\
    .add(Route(route="/users/{id}/calls")
         .for_params(ParamsDecrypterMapper(path_params_keys=['id']))
         .with_get())"""
                "\n"
            ]
        }
    }
}



# Created by Noé Cruz | Zurckz 22 at 01/08/2022
# See https://www.linkedin.com/in/zurckz
import dataclasses
from functools import wraps
from inspect import isclass, signature
from typing import Callable, Optional, Any, Tuple, TypeVar
from zpy.logger import ZLogger
from zpy.utils.values import if_null_get
from zpy.app import zapp_context as ctx
import time

T = TypeVar("T")


@dataclasses.dataclass
class DLazy:
    type: Any
    initializer: Callable[[Any], Any]


class DIContainer:
    """
    Basic Dependency Container
    """

    def __init__(self, timeit: bool = False, x_logger: ZLogger = None,
                 notifier: Callable[[Any, Optional[Any]], None] = None) -> None:
        self.container: dict = dict()
        self.logger: ZLogger = if_null_get(x_logger, ctx().logger)
        self.timeit: bool = timeit
        self.notifier: Callable[[Any, Optional[Any]], None] = notifier
        self.throw_ex = True
        self.error_message_prefix = "Fatal"
        self.max_time_allowed = 5

    def with_notifier(self, notifier):
        self.notifier = notifier

    @classmethod
    def create(cls, timeit: bool = False, logger: ZLogger = None) -> 'DIContainer':
        return cls(timeit, logger)

    def setup(self, init_fn: Callable[['DIContainer'], None]) -> 'DIContainer':
        try:
            te = time.time()
            init_fn(self)
            ts = time.time()
            self.logger.info(f"🚀 Dependencies loaded successfully... {(ts - te) * 1000:2.2f} ms.")
        except Exception as e:
            self.logger.err("Failed to load dependencies...")
            if self.notifier:
                self.notifier(f"{self.error_message_prefix} - Failed to load dependencies: {str(e)}")
            if self.throw_ex:
                raise
        return self

    def __getitem__(self, item: T) -> T:
        return self.get(item)

    def __setitem__(self, key, value):
        self.container[key] = value

    def register(self, x_type: Any, value: Any) -> None:
        """
        Register object or instance of any type in the container.
        @param x_type: Object type to register
        @param value: instance or value to register
        @return: None
        """
        self[x_type] = value

    def factory_register(self, x_type: Any, initializer: Callable[[Any], Any]) -> None:
        """
        Register object or instance of any type in the container using factory strategy.
        This method evaluates the load time, if the load time exceeds the allowed time, the notifier is invoked
        @param x_type: Object type to register
        @param initializer: Loader function
        @return: None
        """
        self[x_type] = self.__timeit__(self.timeit, initializer, x_type, self)

    def lazy_register(self, x_type, initializer: Callable[[Any], Any]) -> None:
        """
        Register object or instance of any type in the container using lazy strategy.
        This method evaluates the load time, if the load time exceeds the allowed time, the notifier is invoked
        @param x_type: Object type to register
        @param initializer: Loader function
        @return: None
        """
        self[x_type] = DLazy(x_type, initializer)

    def get(self, x_type: T, default=None) -> Optional[T]:
        """
        Retrieve object registered in the container
        @param x_type: Object type
        @param default: If object not registered return this value
        @return: Optional object registered or default value
        """
        dependency = self.container.get(x_type, default)

        if isinstance(dependency, DLazy):
            self.container[x_type] = self.__timeit__(self.timeit, dependency.initializer, x_type, self)
            return self.container[x_type]

        return dependency

    def take(self, key: Any, x_type: T, default=None) -> Optional[T]:
        """
        Retrieve object registered in the container
        @param key: key used in registration
        @param x_type: type of value registered
        @param default: if the value registered not exist this value will be returned
        @return: value registered or default
        """
        return self.get(key, default)

    def __timeit__(self, timeit: bool, fn: Any, x_type: Any, *args):
        if not timeit:
            return fn(args[0])
        te = time.time()
        result = fn(args[0])
        ts = time.time()
        taken = ts - te
        self.logger.info(f"Dependency load time: {x_type} :: {taken * 1000:2.2f} ms.")
        if taken >= self.max_time_allowed:
            msg = f"The dependency: {str(x_type)} is exceeding the allowed time. Taken: {taken:2.2f} - Max: {self.max_time_allowed}s."
            self.logger.warn(msg)
            if self.notifier:
                self.notifier(msg)
        return result

    def __getattr__(self, item: T) -> T:
        return self.get(item)


zdi = DIContainer().create()


def populate(initializer, container):
    print(initializer)
    parameters_name: Tuple[str, ...] = tuple(signature(initializer).parameters.keys())
    parameters: Tuple[str, ...] = tuple(signature(initializer).parameters.items())
    print(parameters_name)
    print(parameters)

    @wraps(initializer)
    def _decorated(*args, **kwargs):
        # all arguments were passed
        # if len(args) == len(parameters_name):
        #    return service(*args, **kwargs)

        # if parameters_name == tuple(kwargs.keys()):
        #    return service(**kwargs)

        # all_kwargs = _resolve_kwargs(args, kwargs)
        return initializer(1, 2, 3)

    return _decorated


def inject(container: DIContainer = zdi):
    def _decorator(_service: Any) -> Any:
        if isclass(_service):
            setattr(
                _service,
                "__init__",
                populate(getattr(_service, "__init__"), container)
            )
            return _service

        return _service

    return _decorator



# Created by Noé Cruz | Zurckz 22 at 20/03/2022
# See https://www.linkedin.com/in/zurckz
import dataclasses
import json
from typing import Callable, Optional, Any, Union, Tuple, Dict

from zpy.api.http.errors import ZHttpError
from zpy.api.http.status_codes import HttpStatus
from zpy.app import zapp_context as ctx


class LambdaEventHandler(object):
    """
    Lambda event handler
    """
    _DEFAULT_ERR_MSG = 'An error occurred while processing event data'
    CORS_HEADERS = {
        "Access-Control-Allow-Headers": "Content-Type",
        "Access-Control-Allow-Origin": "*",
        "Access-Control-Allow-Methods": "POST"
    }

    def __init__(self, processor: Optional[Callable[[dict, Any], Any]], strict=False, verbose: bool = False,
                 jsonfy=True, signal_processor: Optional[Callable[[dict, Any], Any]] = None,
                 on_before: Optional[Callable[[dict, Any], Tuple[Dict, Any]]] = None,
                 on_after: Optional[Callable[[dict], Dict]] = None):
        self.processor = processor
        self.strict = strict
        self.verbose = verbose
        self.jsonfy = jsonfy
        self.request_parser = None
        self.response_builder = None
        self.error_notifier = None
        self.send_str_exception = True
        self.starting_process_logger = None
        self.common_fatal_error_msg = f'Fatal :: {self._DEFAULT_ERR_MSG}'
        self.signal_processor = signal_processor
        self.signal_event_key = 'zpy_event_action'
        self.on_before = on_before
        self.on_after = on_after
        self.before: Optional[Callable[[], None]] = None
        self.after: Optional[Callable[[], None]] = None

    def configure(self, response_builder: Callable[[Any, Any], Any] = None,
                  request_parser: Callable[[Any, Any], Any] = None,
                  error_notifier: Callable[[Any, Any], None] = None, send_str_exception: bool = True,
                  starting_process_logger: Callable[[Any, Any], None] = None,
                  common_error_msg: str = None) -> 'LambdaEventHandler':
        self.response_builder = response_builder
        self.request_parser = request_parser
        self.error_notifier = error_notifier
        self.send_str_exception = send_str_exception
        self.starting_process_logger = starting_process_logger
        self.configure_error_msg(self._DEFAULT_ERR_MSG if common_error_msg is None else common_error_msg)
        return self

    def configure_error_msg(self, msg: str) -> None:
        self.common_fatal_error_msg = f'Fatal :: {msg}'

    @classmethod
    def of(cls, processor: Optional[Callable[[dict, dict], Any]], strict=False, verbose=False, jsonfy=True):
        return cls(processor, strict, verbose, jsonfy)

    @staticmethod
    def proxy(event, context, handler):
        return handler.run(event, context)

    def __notify_error(self, record, msg):
        if self.error_notifier:
            self.error_notifier(record, msg)

    def is_wake_up_signal(self, record: dict):
        if not record:
            return False
        prop_key = record.get(self.signal_event_key, None)
        return prop_key == 'WakeUpSignal'

    def run(self, event, context) -> Any:
        logger = ctx().logger
        result = None
        try:
            if self.on_before:
                e, c = self.on_before(event, context)
                event = e
                context = context
            if self.before:
                self.before()

            if self.request_parser:
                event = self.request_parser(event, context)

            if not self.processor:
                msg = "The lambda event processor cant be null."
                self.__notify_error(event, msg)
                return msg
            if type(event) is not dict:
                if self.jsonfy:
                    event = json.loads(event)

            if self.verbose:
                logger.info(f"Starting event processing with data: {event}")
            if self.starting_process_logger:
                self.starting_process_logger(event, context)

            if self.is_wake_up_signal(event):
                result = "Wake up signal received successfully"
                if self.signal_processor:
                    result = self.signal_processor(event, context)
            else:
                result = self.processor(event, context)
            if self.verbose:
                logger.info(result)
        except Exception as e:
            logger.exception("An error occurred processing event", exc_info=e)
            self.__notify_error(event, str(e))
        if self.on_after:
            try:
                result = self.on_after(result)
            except Exception as e:
                logger.exception("An error occurred execute on after hook", exc_info=e)
        if self.after:
            try:
                self.after()
            except Exception as e:
                logger.exception("An error occurred execute on simple after hook", exc_info=e)
        return {"results": result}


@dataclasses.dataclass
class LambdaEventResult(object):
    status: HttpStatus
    payload: dict
    success: bool
    event: str

    def __init__(self, code: HttpStatus, payload: dict, success: bool, event: str = None):
        self.status = code
        self.payload = payload
        self.success = success
        self.event = event


@dataclasses.dataclass
class SuccessEventResult(LambdaEventResult):

    def __init__(self, payload: Union[dict, Any], status: HttpStatus = HttpStatus.SUCCESS):
        LambdaEventResult.__init__(self, status, payload, True, None)


def build_response(status: HttpStatus, payload: dict, success: bool, message: str = None, reason: str = None):
    return {
        "status": {
            "code": status.value[0],
            "message": message if message else status.value[2],
            "details": [reason] if reason else None,
            "status": status.value[1]
        },
        "payload": json.dumps(payload),
        "success": success
    }


def lambda_response(default_status: HttpStatus = HttpStatus.SUCCESS, notifier: Callable[[str], None] = None):
    """
    HTTP Response builder, build response from data provided
    @return: None

    @contact https://www.linkedin.com/in/zurckz
    @author Noé Cruz | Zurck'z 20
    @since 16-05-2020
    """

    def lambda_inner_builder(invoker: Callable):
        def wrapper_lambda_handler(*args, **kwargs):
            try:
                response: Any = invoker(*args, **kwargs)
                if isinstance(response, LambdaEventResult) or isinstance(response, SuccessEventResult):
                    return build_response(response.status, response.payload, response.success)
                return build_response(default_status, response, True)
            except ZHttpError as e:
                msg = f"An error occurred while processing event: Reason: {e.reason}. Message: {e.message}"
                ctx().logger.exception(msg, exc_info=e)
                if notifier:
                    notifier(msg)
                return build_response(e.status, e.metadata, False, e.message, e.reason)
            except Exception as e:
                msg = f"An error occurred while processing event. {str(e)}"
                ctx().logger.exception(msg, exc_info=e)
                if notifier:
                    notifier(msg)
                return build_response(HttpStatus.INTERNAL_SERVER_ERROR, None, False)

        wrapper_lambda_handler.__name__ = invoker.__name__
        return wrapper_lambda_handler

    return lambda_inner_builder


def lambda_api_response(content_type: Optional[str] = 'application/json', default_code: int = 200,
                        cors: bool = True):
    """
    HTTP Response builder, build response from data provided
    @param cors:
    @param default_code:
    @param content_type: Response content type. Default: application/json
    @return: None

    @contact https://www.linkedin.com/in/zurckz
    @author Noé Cruz | Zurck'z 20
    @since 16-05-2020
    """
    extra_headers = {'Content-Type': content_type}

    def lambda_inner_builder(invoker: Callable):
        def wrapper_lambda_handler(*args, **kwargs):
            response: Any = invoker(*args, **kwargs)
            if isinstance(response, tuple) and len(response) == 3:
                data, code, header = response
                if not header:
                    header = extra_headers
                header.update(extra_headers)
            else:
                data = response
                code = default_code
                header = extra_headers

            if cors:
                header.update(LambdaEventHandler.CORS_HEADERS)

            return {
                "statusCode": code,
                "body": json.dumps(data),
                "headers": header
            }

        wrapper_lambda_handler.__name__ = invoker.__name__
        return wrapper_lambda_handler

    return lambda_inner_builder



# Created by Noé Cruz | Zurckz 22 at 20/09/2022
# See https://www.linkedin.com/in/zurckz
from abc import ABC
from enum import Enum
from time import sleep
from typing import Any, Callable, List

from zpy.logger import ZLogger


class ZBuilder(ABC):

    def update(self, property_name: Any, value: Any):
        setattr(self, str(property_name), value)
        return self


class ZRetryPolicyProp(Enum):
    DELAY = 'delay'  # float: Seconds delay on each attempt
    ATTEMPTS = 'attempts'  # int: Attempts number
    LOG_MSG = 'log_message'  # str: Log error message
    DEFAULT_VALUE = 'default_value'  # Any: Default value to return
    RAISE_EXCEPT = 'raise_except'  # bool: True if raise exceptions
    ENABLE_VERBOSE = 'verbose'  # bool: True if print logs

    def __str__(self) -> str:
        return self.value


class ZRetryPolicyBuilder(ZBuilder):

    def __init__(self, *args, **kwargs) -> None:
        ZBuilder.__init__(self)
        self.delay = 0.0
        self.args = args
        self.kwargs = kwargs
        self.attempts = 1
        self.on_excepts = None
        self.on_results = None
        self.logger = None
        self.log_message = "Failed"
        self.default_value = None
        self.raise_except = False
        self.verbose = True

    def on_excepts(self, excepts: List[Exception]):
        self.on_excepts = excepts
        return self

    def on_results(self, results: List[Any]):
        self.on_results = results
        return self

    def with_logger(self, logger):
        self.logger = logger

    def update(self, property_name: ZRetryPolicyProp, value: Any) -> 'ZRetryPolicyBuilder':
        super().update(property_name, value)
        return self

    def build(self) -> 'ZRetryer':
        return ZRetryer(self)


class ZRetryer:

    def __init__(self, policy: ZRetryPolicyBuilder) -> None:
        self.policy = policy

    def with_default(self, value: Any) -> 'ZRetryer':
        self.policy.default_value = value
        return self

    def with_log_message(self, value: Any) -> 'ZRetryer':
        self.policy.log_message = value
        return self

    def __log_message(self, value: Any):
        if self.policy.verbose is True:
            print(value)

    def call(self, task: Callable[[Any], Any], *args, **kwargs):

        max_attempts = self.policy.attempts
        except_detected = None

        while max_attempts >= 0:
            if max_attempts != self.policy.attempts:
                self.__log_message(
                    "Retrying \'{}(*args)\' function invocation".format(task.__name__))
            try:
                x_result = task(*args, **kwargs)
                if self.policy.on_results and x_result in self.policy.on_results:
                    sleep(self.policy.delay)
                    max_attempts = max_attempts - 1
                    continue
                return x_result
            except Exception as e:
                except_detected = e
                if self.policy.on_excepts and type(e) not in self.policy.on_excepts:
                    return self.policy.default_value
            sleep(self.policy.delay)
            max_attempts = max_attempts - 1
        if except_detected:
            if self.policy.log_message:
                self.__log_message(
                    f"{self.policy.log_message} - [RETRYER] Failed {self.policy.attempts} times.")
            else:
                self.__log_message(
                    f"[RETRYER] Failed {self.policy.attempts} times.")
            if self.policy.raise_except is True:
                raise except_detected
        return self.policy.default_value


def retry(attempts: int = 1, delayed: float = 0.5, on_excepts: List[Exception] = None, on_results: List[Any] = None,
          default_result=None, log_message: str = None, logger: ZLogger = None, raise_except=False) -> Any:
    """
    Function decorator for retry function execution
    @param attempts: Number of attempts to retry.
    @param delayed: Time delay between each attempt
    @param on_excepts: Exceptions list to retry function execution
    @param on_results: Exceptions results to retry function execution
    @param default_result: Default value returned if the function failed
    @param log_message: Message for log
    @param logger: ZLogger instance
    @param raise_except: True if you need raise exception otherwise default value returned
    @return: Function result or default value
    """

    def decorator(fun):
        def wrapper(*args, **kwargs):
            max_attempts = attempts
            except_detected = None
            while max_attempts >= 0:
                if logger and max_attempts != attempts:
                    logger.info("Retrying \'{}(*args)\' function invocation".format(fun.__name__))
                try:
                    result = fun(*args, **kwargs)
                    if on_results and result in on_results:
                        sleep(delayed)
                        max_attempts = max_attempts - 1
                        continue
                    return result
                except Exception as e:
                    except_detected = e
                    if on_excepts and type(e) not in on_excepts:
                        return default_result
                sleep(delayed)
                max_attempts = max_attempts - 1
            if except_detected:
                if log_message and logger:
                    logger.err(f"{log_message} - [RETRYER] Failed {attempts} times.")
                if raise_except is True:
                    raise except_detected
            return default_result

        return wrapper

    return decorator



from typing import Any, List, Callable, Optional, TypeVar

T = TypeVar('T')


def first(collection: List[T], fn: Callable[[T], Any] = None, default: Optional[T] = None) -> T:
    value = collection[0] if collection else default
    return fn(value) if fn and value else value


def last(collection: List[T], fn: Callable[[T], Any] = None, default: Optional[T] = None) -> T:
    value = collection[-1] if collection else default
    return fn(value) if fn and value else value


def find(any_list: List[Any], x_filter: Callable, default_value: Any = None) -> Optional[object]:
    """
    Find item in provided list according filter
    @param default_value: if not found element will be returned
    @param any_list: List of any values
    @param x_filter: Filter to apply the search
    @return: element that filter return true otherwise None
    """
    for x in any_list:
        if x_filter(x):
            return x
    return default_value


def walk_and_apply(dict_value: dict, func: Callable[[str, Any], Any], skip_keys: List[str] = None):
    """
    Iterate over dict and apply function for override value.
    """
    if not skip_keys:
        skip_keys = []

    for k in dict_value.keys():
        if isinstance(dict_value[k], dict):
            walk_and_apply(dict_value[k], func)
            continue
        if k in skip_keys:
            continue
        dict_value[k] = func(k, dict_value[k])



# Created by Noé Cruz | Zurckz 22 at 11/09/2022
# See https://www.linkedin.com/in/zurckz
from logging import INFO, WARN, ERROR
from typing import Optional

from zpy.logger import ZLogger
from zpy.app import zapp_context as ctx, ZContext
from zpy.utils.values import if_null_get
from zpy.logger.structured import StructuredLog


class Loggable:

    def __init__(self, logger: Optional[ZLogger] = None):
        self.logger: ZLogger = if_null_get(logger, ctx().logger)
        self.app = ctx().app_name
        self.context: ZContext = ctx()

    def log_exception(self, value: any, shippable: bool = False, **kwargs):
        self.logger.exception(value, shippable, **kwargs)

    def log_error(self, value: any, exc_info: any, shippable: bool = False):
        self.logger.err(value, shippable, exc_info=exc_info)

    def log_info(self, value: any, shippable: bool = False, **kwargs):
        self.logger.info(value, shippable, **kwargs)

    def log_warn(self, value: any, shippable: bool = False, **kwargs):
        self.logger.warn(value, shippable, **kwargs)

    def log_struct_info(self, message: str, flow: str = None, description: str = None, meta: Optional[dict] = None,
                        **kwargs) -> StructuredLog:
        _flow = if_null_get(flow, self.context.get_flow())
        log = StructuredLog(_flow, message, self.app, description, INFO, meta, **kwargs)
        self.logger.info(log)
        return log

    def log_struct_warn(self, message: str, flow: str = None, description: str = None, meta: Optional[dict] = None,
                        **kwargs) -> 'StructuredLog':
        log = StructuredLog(if_null_get(flow, self.context.get_flow()), message, self.app, description, WARN, meta,
                            **kwargs)
        self.logger.warn(log)
        return log

    def log_struct_error(self, message: str, flow: str = None, description: str = None, meta: Optional[dict] = None,
                         **kwargs) -> 'StructuredLog':
        log = StructuredLog(if_null_get(flow, self.context.get_flow()), message, self.app, description, ERROR, meta,
                            **kwargs)
        self.logger.err(log)
        return log



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zpy-api-core",
    version="1.6.0",
    author="Noé Cruz | linkedin.com/in/zurckz/",
    author_email="zurckz.services@gmail.com",
    description="Helper layer for apis development with Aws, Python and Flask",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NoeCruzMW",
    packages=setuptools.find_packages(),
    install_requires=[
        "Flask",
        "flask_cors",
        "marshmallow",
        "marshmallow_objects",
        "aws-lambda-wsgi",
        "requests",
        "Click",
        "python-dateutil"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    py_modules=['zpy.cli'],
    entry_points={
        'console_scripts': [
            'zpy = zpy.cli:cli',
        ],
    },
    python_requires=">=3.6",
)



# Created by Noé Cruz | Zurckz 22 at 06/11/2022
# See https://www.linkedin.com/in/zurckz
from typing import TypeVar, Generic, Callable, Any

T = TypeVar('T')
R = TypeVar('R')


class Ok(Generic[T]):
    def __init__(self, value: T, meta=None):
        self.value = value
        self.meta = meta

    @staticmethod
    def of(value: T, meta=None) -> 'Result':
        return Result(Ok(value, meta), Failure.none())

    @classmethod
    def none(cls):
        return cls(None)


class Failure(Generic[T]):

    def __init__(self, value: T, meta=None):
        self.value = value
        self.meta = meta

    @classmethod
    def none(cls):
        return cls(None)

    @staticmethod
    def of(value: T, meta=None) -> 'Result':
        return Result(Ok.none(), Failure(value, meta))


class Result(Generic[T, R]):

    def __init__(self, ok: Ok[T], failure: Failure[R]):
        self.ok = ok
        self.failure = failure

    def unwrap(self, default=None) -> T:
        if self.ok and self.ok.value is not None:
            return self.ok.value
        return default

    def is_ok(self) -> bool:
        return self.ok and self.ok.value is not None

    def unfold(self, ok: Callable[[T], Any], fail: Callable[[R], Any]) -> Any:
        if self.is_ok():
            return ok(self.ok.value)
        return fail(self.failure.value)

    def metadata(self) -> Any:
        if self.is_ok():
            return self.ok.meta
        return self.failure.meta



import json
from abc import ABC, abstractmethod
from datetime import timedelta
from functools import wraps
from timeit import default_timer as timer
from typing import Any, Optional, List, Dict, Tuple
from typing import Callable

from zpy.api.http.errors import ZHttpError
from zpy.api.http.response import _useAwsRequestId
from zpy.api.http.status_codes import HttpMethods
from zpy.app import zapp_context as api
from zpy.app import zapp_context as ctx
from zpy.containers import shared_container
from zpy.logger import zL
from zpy.utils.values import if_null_get

DEFAULT_RESPONSE = {'statusCode': '500', 'headers': {'Content-Type': 'application/json', 'Content-Length': '120',
                                                     'Access-Control-Allow-Origin': '*'}, 'isBase64Encoded': False,
                    'body': '{"code":"INTERNAL SERVER ERROR","details":["The execution of a hook failed before reaching the main process."],"message":"The process could not be completed due to a semantics error."}'}


class AWSEventStep(ABC):
    """
     Event Step
    """
    name: str
    raise_fails: Any
    response: dict

    def __init__(self, name: str, raise_fails: bool = True, response: dict = None):
        self.name = name
        self.raise_fails = raise_fails
        self.response = if_null_get(response, DEFAULT_RESPONSE)

    @abstractmethod
    def before(self, event: dict, contex: dict, *args, **kwargs) -> Tuple[Dict, Dict]:
        ...

    @abstractmethod
    def after(self, response: dict, *args, **kwargs) -> dict:
        """"""
        ...


class EventMapper(ABC):
    @abstractmethod
    def configure(self, *args, **kwargs) -> None:
        pass

    @abstractmethod
    def execute(self, *args, **kwargs) -> Any:
        pass


class RequestEventMapper(EventMapper):
    @abstractmethod
    def configure(self, event: dict, context: Any, *args, **kwargs) -> None:
        pass

    @abstractmethod
    def execute(self, event: dict, context: Any, *args, **kwargs) -> Tuple[Any, Any]:
        pass


class ResponseEventMapper(EventMapper):
    @abstractmethod
    def configure(self, response: dict, *args, **kwargs) -> None:
        pass

    @abstractmethod
    def execute(self, response: dict, *args, **kwargs) -> Any:
        pass


class RouteEventMapper:
    """

    """

    def __init__(self, route: str, http_method: HttpMethods = HttpMethods.GET,
                 full_request_mapper: Optional[RequestEventMapper] = None,
                 request: EventMapper = None,
                 response: ResponseEventMapper = None,
                 headers: EventMapper = None,
                 query: EventMapper = None,
                 path: EventMapper = None, *args, **kwargs) -> None:
        self.route = route
        self.http_method = http_method
        self.full_request_mapper: RequestEventMapper = full_request_mapper
        self.request_mappers: Dict[str, EventMapper] = {
            'request': request,
            'headers': headers,
            'query_params': query,
            'path_params': path
        }
        self.response_mapper: ResponseEventMapper = response
        self.extra_args = args
        self.extra_kwargs = kwargs

    def for_request(self, mapper: RequestEventMapper):
        self.request_mappers['request'] = mapper
        return self

    def for_response(self, mapper: ResponseEventMapper):
        self.response_mapper = mapper
        return self

    def for_headers(self, mapper: RequestEventMapper):
        self.request_mappers['headers'] = mapper
        return self

    def for_params(self, mapper: RequestEventMapper):
        self.request_mappers['query_params'] = mapper
        return self

    def for_path(self, mapper: RequestEventMapper):
        self.request_mappers['path_params'] = mapper
        return self

    def with_method(self, method: HttpMethods):
        self.http_method = method
        return self

    def with_meta(self, key: str, value: Any):
        self.extra_kwargs[key] = value
        return self

    def with_post(self):
        """
        Configure POST http method
        @return:
        """
        self.http_method = HttpMethods.POST
        return self

    def with_get(self):
        """
        Configure GET http method
        @return:
        """
        self.http_method = HttpMethods.GET
        return self

    def with_put(self):
        """
        Configure GET http method
        @return:
        """
        self.http_method = HttpMethods.PUT
        return self

    def with_patch(self):
        """
        Configure PATCH http method
        @return:
        """
        self.http_method = HttpMethods.PATCH
        return self

    def with_delete(self):
        """
        Configure DELETE http method
        @return:
        """
        self.http_method = HttpMethods.DELETE
        return self

    def configure_request_mappers(self, event: Dict, context: Any, *args, **kwargs):
        if self.full_request_mapper:
            self.full_request_mapper.configure(event, context, *(args + self.extra_args),
                                               **{**kwargs, **self.extra_kwargs})

        for k in self.request_mappers:
            if self.request_mappers[k] and self.request_mappers[k] != self.full_request_mapper:
                self.request_mappers[k].configure(event, context, *(args + self.extra_args),
                                                  **{**kwargs, **self.extra_kwargs})

    def configure_response_mappers(self, response: Any, *args, **kwargs):
        if self.response_mapper:
            self.response_mapper.configure(response, *(args + self.extra_args),
                                           **{**kwargs, **self.extra_kwargs})

    @classmethod
    def from_dict(cls, config: Dict[str, Dict[str, Any]], key: str = None):
        if not config:
            raise ValueError('Config cannot be null.')
        keys: List[str] = list(config.keys())
        if not keys:
            raise ValueError('Config should be a key')

        str_route: str = if_null_get(key, keys[0])

        def get_map(x):
            return x.get('mapper', None)

        def get(x, k) -> Dict:
            return x.get(k, {})

        spec = get(config, str_route)
        default_mapper = get(spec, 'default').get('mapper', None)

        return cls(route=str_route, full_request_mapper=default_mapper, request=get_map(get(spec, 'request')),
                   response=get_map(get(spec, 'response')), headers=get_map(get(spec, 'headers')),
                   query=get_map(get(spec, 'query_params')), path=get_map(get(spec, 'path_params')))


class RouteEventMapperManager(AWSEventStep):

    def __init__(self, configs: List[RouteEventMapper] = None, name: str = 'RouteEventMapperManager',
                 strict: bool = False, initializer: Callable[[Any, Any, Any, Any], None] = None, *args, **kwargs):
        super().__init__(name)
        if not configs:
            configs = []
        self.configs: Dict[str, RouteEventMapper] = {f'{k.route}:{k.http_method.value}': k for k in configs}
        self.current: Optional[RouteEventMapper] = None
        self.config_found = False
        self.strict = strict
        self.extra_args = args
        self.extra_kwargs = kwargs
        self.initializer = initializer

    def add(self, route_config: RouteEventMapper) -> 'RouteEventMapperManager':
        self.configs[f'{route_config.route}:{route_config.http_method.value}'] = route_config
        return self

    def add_meta(self, key: str, value: Any) -> 'RouteEventMapperManager':
        self.extra_kwargs[key] = value
        return self

    @classmethod
    def from_dict(cls, configs: Dict, name: str):
        configurations = [RouteEventMapper.from_dict(configs, k) for k in configs.keys()]
        return cls(configurations, name)

    def before(self, event: dict, context: dict, *args, **kwargs) -> Tuple[Dict, Any]:
        if not event:
            return event, context
        if 'resource' not in event or 'httpMethod' not in event:
            return event, context
        current_route = f'{event["resource"]}:{event["httpMethod"]}'
        self.config_found = False
        if current_route in self.configs:
            if self.initializer:
                self.initializer(event, context, *self.extra_args, **self.extra_kwargs)
            self.current = self.configs[current_route]
            self.current.configure_request_mappers(event, context, *self.extra_args, **self.extra_kwargs)
            self.config_found = True
            if self.current.full_request_mapper:
                event, context = self.current.full_request_mapper.execute(event=event, context=context,
                                                                          *self.extra_args,
                                                                          **self.extra_kwargs)
            for x_mapper in self.current.request_mappers.values():
                if x_mapper is not None:
                    event, context = x_mapper.execute(event, context=context, *self.extra_args,
                                                      **self.extra_kwargs)
        return event, context

    def after(self, response: dict, *args, **kwargs) -> dict:
        if self.config_found and self.current.response_mapper is not None:
            self.current.configure_response_mappers(response, *self.extra_args, **self.extra_kwargs)
            return self.current.response_mapper.execute(response=response, *self.extra_args, **self.extra_kwargs)
        return response


def prepare_event_pipe_exception(e: ZHttpError) -> dict:
    body = {"code": f"{e.get_str_code()}", "details": e.details,
            "message": f"{e.get_message()}"}
    body = json.dumps(body)
    return {'statusCode': f'{e.get_http_code()}',
            'headers': {'Content-Type': 'application/json', 'Content-Length': len(body),
                        'Access-Control-Allow-Origin': '*'}, 'isBase64Encoded': False,
            'body': body}


def lambda_event_pipe(event: dict, context: dict,
                      processor: Callable[[Dict, Dict], Dict],
                      steps: Optional[List[AWSEventStep]] = None):
    logger = ctx().logger
    response = None
    if not steps:
        steps = []
    for mw in steps:
        try:
            event, context = mw.before(event, context)
        except ZHttpError as e:
            logger.exception(f"An error occurred when execute before: {mw.name}. ", exc_info=e)
            return prepare_event_pipe_exception(e)
        except Exception as e:
            logger.exception(f"An error occurred when execute before: {mw.name}. ", exc_info=e)
            if mw.raise_fails:
                return mw.response
    try:
        response = processor(event, context)
    except Exception as e:
        logger.exception(f"An error occurred when execute processor... ", exc_info=e)
        return DEFAULT_RESPONSE

    for mw in reversed(steps):
        try:
            response = mw.after(response)
        except ZHttpError as e:
            logger.exception(f"An error occurred when execute after: {mw.name}. ", exc_info=e)
            return prepare_event_pipe_exception(e)
        except Exception as e:
            logger.exception(f"An error occurred when execute after: {mw.name}. ", exc_info=e)
            if mw.raise_fails:
                return mw.response
    return response


class LambdaEventPipe:

    def __init__(self, event: dict, context: Any):
        self.event = event
        self.context = context
        self.steps = []

    def add(self, step: AWSEventStep) -> 'LambdaEventPipe':
        self.steps.append(step)
        return self

    def run(self, event_processor: Callable[[Dict, Any], Dict]) -> dict:
        return lambda_event_pipe(self.event, self.context, event_processor, self.steps)


def store_request_id(context) -> Optional[str]:
    """Extract aws request id from context

    Args:
        context ([type]): Lambda context
    """
    try:
        shared_container["aws_request_id"] = context.aws_request_id
        return context.aws_request_id
    except Exception as e:
        zL.ex("An error occurred while extracting aws request id", exc_info=e)
        return None


def event_processors(storage: dict, use_id: bool, logs: bool, send_logs: bool, *args, **kwargs):
    """Lambda event processors

    Args:
        @param storage:
        @param use_id:
        @param logs:
        @param send_logs:
    """
    try:
        if len(args) >= 2:
            event = args[0]
            storage['request'] = event
            if logs:
                api().logger.info(f"Request: {event}", shippable=send_logs)
            if _useAwsRequestId or use_id:
                storage['request_id'] = store_request_id(args[1])

        else:
            if "event" in kwargs:
                storage['request'] = kwargs['event']
                if logs:
                    api().logger.info(f"Request: {kwargs['event']}", shippable=send_logs)
            if "context" in kwargs:
                if _useAwsRequestId or use_id:
                    storage['request_id'] = store_request_id(args[1])
    except Exception as e:
        api().logger.ex("An error occurred while processing event!", exc_info=e)


def aws_lambda(logs: bool = True, save_id: bool = False, measure_time: bool = True, send_logs: bool = False,
               event_sender: Optional[Callable[[dict], Any]] = None):
    """Lambda Handler

    Args:
        @param event_sender:
        @param logs: (bool, optional): Logging request and response. Defaults to False.
        @param save_id: (bool, optional): Register aws lambda request id. Defaults to True.
        @param measure_time: (bool, optional): Measure elapsed execution time. Defaults to True.
        @param send_logs: Send event logs by log sender configured
    """
    api().release_logger()
    event = {'request_id': '-'}

    def callable_fn(invoker: Callable):
        @wraps(invoker)
        def wrapper(*args, **kwargs):
            event_processors(event, save_id, logs, send_logs, *args, **kwargs)
            start = 0.0
            if if_null_get(measure_time, False):
                start = timer()
            result = invoker(*args, **kwargs)
            event['response'] = result
            if logs:
                api().logger.info(f"Response: {result}", shippable=send_logs)
            if if_null_get(measure_time, False):
                end = timer()
                api().logger.info(f"Elapsed execution time: {timedelta(seconds=end - start)}", shippable=send_logs)
            if event_sender:
                event_sender(event)
            return result

        return wrapper

    return callable_fn
